# SPDX-FileCopyrightText: 2020 2020
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import absolute_import
from builtins import object
from . import arf_consts as ac
import os
import sys
from os import path as op
from os import linesep
from . import alert_actions_exceptions as aae
from munch import Munch
from mako.template import Template
from mako.lookup import TemplateLookup
from lxml import etree, html
from re import search
from .alert_actions_template import AlertActionsTemplateMgr
from .alert_actions_helper import write_file


class AlertHtmlBase(object):

    def __init__(self, input_setting=None, package_path=None, logger=None):
        self._all_setting = input_setting
        self._logger = logger
        self._package_path = package_path
        self._current_alert = None

    def get_alert_html_name(self):
        return self._current_alert[ac.SHORT_NAME] + ".html"

    def get_alert_html_path(self):
        if not self._package_path:
            return None

        html_path = op.join(self._package_path, "default", "data", "ui",
                            "alerts")
        if not op.exists(html_path):
            os.makedirs(html_path)

        return op.join(html_path,
                       self.get_alert_html_name())


class AlertHtmlGenerator(AlertHtmlBase):
    DEFAULT_TEMPLATE_HTML = "mod_alert.html.template"
    DEFAULT_HOME_HTML = "default.html"

    def __init__(self, input_setting=None, package_path=None, logger=None,
                 html_template=None,
                 html_home=None,
                 html_theme=None):
        super(AlertHtmlGenerator, self).__init__(input_setting, package_path,
                                                 logger)
        if not input_setting or not logger:
            msg = 'required_args="input_setting, logger"'
            raise aae.AlertActionsInValidArgs(msg)

        self._alert_actions_setting = input_setting[ac.MODULAR_ALERTS]
        self._template = None
        self._html_template = html_template or \
            AlertHtmlGenerator.DEFAULT_TEMPLATE_HTML
        self._html_home = html_home or AlertHtmlGenerator.DEFAULT_HOME_HTML
        self._temp_obj = AlertActionsTemplateMgr(html_theme=html_theme)
        self._html_theme = self._temp_obj.get_html_lookup_dir()
        self._logger.info('html_theme="%s" html_template="%s", html_home="%s"',
                          self._html_theme,
                          self._html_template,
                          self._html_home)

    def handle_one_alert(self, one_alert_setting):
        self._current_alert = one_alert_setting
        alert_obj = Munch.fromDict(one_alert_setting)
        final_form = self._template.render(mod_alert=alert_obj,
                                           home_page=self._html_home)
        final_form = html.fromstring(final_form)
# Checking python version before converting and encoding XML Tree to string.
        if sys.version_info < (3, 0):
            final_string = etree.tostring(final_form, encoding='utf-8',
                                          pretty_print=True)
        else:
            final_string = etree.tostring(
                final_form, encoding='utf-8', pretty_print=True)
        text = linesep.join(
            [s for s in final_string.decode('utf-8').splitlines() if not search(r'^\s*$', s)])

        self._logger.debug('operation="Write", object_type="File", object="%s"',
                           self.get_alert_html_path())

        write_file(self.get_alert_html_name(),
                   self.get_alert_html_path(),
                   text,
                   self._logger)

    def handle(self):
        self._logger.info("html_theme=%s", self._html_theme)
        tmp_lookup = TemplateLookup(directories=[self._html_theme])
        template_text = None
        template_path = self._html_template
        if not op.isabs(self._html_template):
            template_path = op.join(self._temp_obj.get_template_dir(),
                                    self._html_template)

        self._logger.debug("Reading template_file=%s", template_path)
        with open(template_path, 'r') as tp:
            template_text = tp.read()

        self._template = Template(text=template_text,
                                  lookup=tmp_lookup)

        self._logger.info("Start to generate alert actions html files")
        for alert in self._alert_actions_setting:
            self.handle_one_alert(alert)
        self._logger.info("Finished generating alert actions html files")


def generate_alert_actions_html_files(input_setting=None, package_path=None,
                                      logger=None,
                                      html_setting=None):
    html_template = None
    html_home = None
    html_theme = None
    if html_setting:
        html_template = html_setting.get("html_template")
        html_home = html_setting.get("html_home")
        html_theme = html_setting.get("html_theme")
    html_gen = AlertHtmlGenerator(input_setting=input_setting,
                                  package_path=package_path,
                                  logger=logger,
                                  html_template=html_template,
                                  html_home=html_home,
                                  html_theme=html_theme
                                  )
    html_gen.handle()
    return None



# SPDX-FileCopyrightText: 2020 2020
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import absolute_import
import solnlib.log as log
from . import builder_constant
import logging
import threading
import copy

from alert_utils.alert_utils_common.metric_collector import metric_util

LOGS = {
    "validation": "ta_builder_validation",
    "field_extraction_validator": "ta_builder_validation",
    "app_cert_validator": "ta_builder_validation",
    "meta_manager": "ta_builder",
    "sourcetype": "ta_builder",
    "utility": "ta_builder",
    "app": "ta_builder",
    "generator": "ta_builder",
    "cim_builder": "ta_builder",
    "input_builder": "ta_builder",
    "setup_builder": "ta_builder",
    "field_extraction_builder": "ta_builder",
    "alert_builder": "ta_builder",
    "global_settings_builder": "ta_builder",
    "modular_alert": "ta_builder",
    "modular_alert_testing": "ta_builder",
    "package_add_on": "package_add_on"
}

g_log_levels = {}
g_loggers = {}
g_logger_lock = threading.Lock()

DEFAULT_LOG_LEVEL = logging.INFO

# set the context of the log
log.Logs.set_context(namespace=builder_constant.ADDON_BUILDER_APP_NAME)

@metric_util.function_run_time(tags=['tab_logger'])
def get_package_add_on_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("package_add_on"), level)

@metric_util.function_run_time(tags=['tab_logger'])
def get_field_extraction_builder_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("field_extraction_builder"), level)


@metric_util.function_run_time(tags=['tab_logger'])
def get_builder_util_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("utility"), level)


@metric_util.function_run_time(tags=['tab_logger'])
def get_meta_manager_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("meta_manager"), level)


@metric_util.function_run_time(tags=['tab_logger'])
def get_validation_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("validation"), level)


@metric_util.function_run_time(tags=['tab_logger'])
def get_field_extraction_validator_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("field_extraction_validator"), level)


@metric_util.function_run_time(tags=['tab_logger'])
def get_app_cert_validator_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("app_cert_validator"), level)


@metric_util.function_run_time(tags=['tab_logger'])
def get_sourcetype_builder_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("sourcetype"), level)


@metric_util.function_run_time(tags=['tab_logger'])
def get_app_instance_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("app"), level)


@metric_util.function_run_time(tags=['tab_logger'])
def get_generator_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("generator"), level)


@metric_util.function_run_time(tags=['tab_logger'])
def get_cim_builder_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("cim_builder"), level)


@metric_util.function_run_time(tags=['tab_logger'])
def get_input_builder_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("input_builder"), level)


@metric_util.function_run_time(tags=['tab_logger'])
def get_setup_builder_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("setup_builder"), level)

@metric_util.function_run_time(tags=['tab_logger'])
def get_modular_alert_builder_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("modular_alert"), level)

@metric_util.function_run_time(tags=['tab_logger'])
def get_modular_alert_testing_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("modular_alert_testing"), level)

@metric_util.function_run_time(tags=['tab_logger'])
def set_log_level(level, name=None):
    if name:
        logger = _get_logger(name)
        logger.setLevel(level)
    else:
        for name in list(LOGS.values()):
            logger = _get_logger(name)
            logger.setLevel(level)


@metric_util.function_run_time(tags=['tab_logger'])
def _get_logger(name, level=logging.INFO):
    with g_logger_lock:
        l = None
        if name in g_loggers:
            l = g_loggers[name]
        else:
            l = log.Logs().get_logger(name)
            g_loggers[name] = l
        lv = g_log_levels.get(name, 10000)
        if level < lv:
            # logging.DEBUG < INFO < WARN < ERROR
            l.setLevel(level)
            g_log_levels[name] = level
        return l


@metric_util.function_run_time(tags=['tab_logger'])
def get_alert_builder_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("alert_builder"), level)


@metric_util.function_run_time(tags=['tab_logger'])
def get_global_settings_builder_logger(level=DEFAULT_LOG_LEVEL):
    return _get_logger(LOGS.get("global_settings_builder"), level)

HIDDEN_FIELDS = ['password', 'credential', 'value']
def hide_sensitive_field(raw):
    if isinstance(raw, dict):
        cloned = dict(raw)
        for k in list(raw.keys()):
            lower_key = k.lower()
            if any([i in lower_key for i in HIDDEN_FIELDS]):
                cloned[k] = '*******'
            else:
                cloned[k] = hide_sensitive_field(cloned[k])
        return cloned
    elif isinstance(raw, list) or isinstance(raw, tuple):
        cloned = [hide_sensitive_field(i) for i in raw]
        return cloned
    return raw



# SPDX-FileCopyrightText: 2020 2020
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import absolute_import

from .base import RestEntityBuilder, RestEndpointBuilder


class SingleModelEntityBuilder(RestEntityBuilder):
    def __init__(self, name, fields, **kwargs):
        super(SingleModelEntityBuilder, self).__init__(name, fields, **kwargs)

    @property
    def name_spec(self):
        return "<name>"

    @property
    def name_default(self):
        return "default"

    @property
    def name_rh(self):
        return ""


class SingleModelEndpointBuilder(RestEndpointBuilder):

    _rh_template = """
from splunktaucclib.rest_handler.endpoint import (
    field,
    validator,
    RestModel,
    SingleModel,
)
from splunktaucclib.rest_handler import admin_external, util
from {handler_module} import {handler_name}
import logging

util.remove_http_proxy_env_vars()

{entity}

endpoint = SingleModel(
    '{conf_name}',
    model,
    config_name='{config_name}'
)


if __name__ == '__main__':
    logging.getLogger().addHandler(logging.NullHandler())
    admin_external.handle(
        endpoint,
        handler={handler_name},
    )
"""

    def actions(self):
        return ["edit", "list", "remove", "create"]

    def generate_rh(self, handler):
        entity = self._entities[0]
        return self._rh_template.format(
            handler_module=handler.module,
            handler_name=handler.name,
            entity=entity.generate_rh(),
            conf_name=self.conf_name,
            config_name=self._name,
        )



# SPDX-FileCopyrightText: 2020 2020
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import absolute_import

from builtins import object
from .base import quote_string, indent


class RestFieldBuilder(object):

    _kv_template = '{name} = {value}'
    _rh_template = """field.RestField(
    {name},
    required={required},
    encrypted={encrypted},
    default={default},
    validator={validator}
)"""

    def __init__(self, field, validator):
        self._field = field
        self._validator = validator

    def generate_spec(self):
        return self._kv_template.format(
            name=self._field.name,
            value='',
        )

    def _indent_validator(self):
        validator = indent(self._validator)
        return validator[4:]

    def generate_rh(self):
        return self._rh_template.format(
            name=quote_string(self._field.name),
            required=self._field.required,
            encrypted=self._field.encrypted,
            default=quote_string(self._field.default),
            validator=self._indent_validator(),
        )



# SPDX-FileCopyrightText: 2020 2020
#
# SPDX-License-Identifier: Apache-2.0

class AlertParameterInputSettingNotImplemented(Exception):
    pass


class AlertIconPathInvalid(Exception):
    pass


class AlertIconFileDoesNotExist(Exception):
    pass


class AlertActionsInValidArgs(Exception):
    pass


class AlertActionsNotSupportedFeature(Exception):
    pass


class AlertActionsDoWorkNotImplemented(Exception):
    pass


class AlertActionsFailedToLoadInputSettingJson(Exception):
    pass


class AlertTestSettingFileDoesNotExist(Exception):
    pass


class AlertTestInputUnsupportedMode(Exception):
    pass


class AlertTestJsonFileLoadFailure(Exception):
    pass


class AlertTestWritingFileFailure(Exception):
    pass


class AlertSearchCreatingFailure(Exception):
    pass


class AlertTestBuildTAFailure(Exception):
    pass


class AlertTestKillingSubprocessFailure(Exception):
    pass


class AlertTestSubprocessTimeoutFailure(Exception):
    pass


class AlertTestCodeFileNotExistFailure(Exception):
    pass


class AlertCleaningFormatFailure(Exception):
    pass



# SPDX-FileCopyrightText: 2020 2020
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import absolute_import

from past.builtins import basestring
from splunktaucclib.rest_handler.schema import RestSchema
from io import StringIO
from builtins import object
from six import string_types as basestring
import six
from future import standard_library

standard_library.install_aliases()

__all__ = [
    "RestEntityBuilder",
    "RestEndpointBuilder",
    "quote_string",
    "indent",
]


class RestEntityBuilder(object):

    _title_template = "[{}]"
    _rh_template = """
fields{name_rh} = [
{fields}
]
model{name_rh} = RestModel(fields{name_rh}, name={name})
"""
    _disabled_feild_template = """
field.RestField(
    'disabled',
    required=False,
    validator=None
)
"""

    def __init__(self, name, fields, **kwargs):
        self._name = name
        self._fields = fields
        self._conf_name = kwargs.get("conf_name")

    @property
    def name(self):
        return self._name

    @property
    def name_spec(self):
        raise NotImplementedError()

    @property
    def name_default(self):
        raise NotImplementedError()

    @property
    def name_rh(self):
        raise NotImplementedError()

    def generate_spec(self, omit_kv_pairs=False):
        title = self._title_template.format(self.name_spec)
        if omit_kv_pairs:
            return title
        lines = [field.generate_spec() for field in self._fields]
        lines.insert(0, title)
        return "\n".join(lines)

    def generate_rh(self):
        fields = []
        for field in self._fields:
            field_line = field.generate_rh()
            fields.append(field_line)
        # add disabled field for data input
        entity_builder = self.__class__.__name__
        if (
            entity_builder == "DataInputEntityBuilder"
            or entity_builder == "SingleModelEntityBuilder"
            and self._conf_name
        ):
            fields.append(self._disabled_feild_template)
        fields_lines = ", \n".join(fields)
        return self._rh_template.format(
            fields=indent(fields_lines),
            name_rh=self.name_rh,
            name=quote_string(self._name),
        )


class RestEndpointBuilder(object):
    def __init__(self, name, namespace, **kwargs):
        self._name = name
        self._namespace = namespace
        self._entities = []
        self._conf_name = (
            kwargs.get("conf_name")
            if kwargs.get("conf_name") is not None
            else self.name.lower()
        )
        if kwargs.get("rest_handler_name") is not None:
            self._rest_handler_name = kwargs.get("rest_handler_name")
        else:
            self._rest_handler_name = "{}_rh_{}".format(self._namespace, self._name)

    @property
    def name(self):
        return RestSchema.endpoint_name(self._name, self._namespace)

    @property
    def namespace(self):
        return self._namespace

    @property
    def conf_name(self):
        return self._conf_name

    @property
    def rh_name(self):
        return self._rest_handler_name

    @property
    def entities(self):
        return self._entities

    def add_entity(self, entity):
        self._entities.append(entity)

    def actions(self):
        raise NotImplementedError()

    def generate_spec(self):
        specs = [entity.generate_spec() for entity in self._entities]
        return "\n\n".join(specs)

    def generate_default_conf(self):
        specs = [entity.generate_spec(True) for entity in self._entities]
        return "\n\n".join(specs)

    def generate_rh(self, handler):
        raise NotImplementedError()


def quote_string(value):
    """
    Quote a string
    :param value:
    :return:
    """
    if isinstance(value, basestring):
        return "'%s'" % value
    else:
        return value


def quote_regex(value):
    """
    Quote a regex
    :param value:
    :return:
    """
    if isinstance(value, basestring):
        return '"""%s"""' % value
    else:
        return value


def indent(lines, spaces=1):
    """
    Indent code block.

    :param lines:
    :type lines: str
    :param spaces: times of four
    :return:
    """
    string_io = StringIO(six.text_type(lines))
    indentation = spaces * 4
    prefix = " " * indentation
    lines = []
    for line in string_io:
        if line != "\n":
            line = prefix + line
        lines.append(line)
    return "".join(lines)



# SPDX-FileCopyrightText: 2020 2020
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import absolute_import
import re
import shutil
import os
from .alert_actions_conf_gen import generate_alert_actions_conf
from .alert_actions_html_gen import generate_alert_actions_html_files
from .alert_actions_py_gen import generate_alert_actions_py_files
import traceback


cache_path = {}


def check_file_name(file_name, env):
    # file_name is string, read from setting file
    # this function replaces all ${short_name} and other to its real value
    if file_name in cache_path:
        return cache_path[file_name]

    search = re.findall(r"\${\!?([\w\-\.]+)}", file_name, re.MULTILINE)
    if not search:
        cache_path[file_name] = file_name
        return file_name

    new_str = file_name
    for gp in search:
        if gp in env:
            new_str = new_str.replace("${%s}" % gp, env[gp], re.MULTILINE)
            new_str = new_str.replace("${!%s}" % gp,
                                      re.sub("[^\w]+", "_", env[gp].lower()),
                                      re.MULTILINE)

    # Disable the cache to avoid conflict
    # cache_path[file_name] = new_str
    return new_str


def check_file_list(dirName, file_list, env):
    ret = []
    for dname in file_list:
        new_dname = check_file_name(dname, env)
        if new_dname != dname:
            ret.append((check_file_name(dirName, env), dname, new_dname))

    return ret


def prepare_ta_directory_tree(src, dest, logger, envs):
    """
    If dest doesn't exist, then generate a new TA directory tree.
    If dest exists, then merge with the new one
    """
    output_dir = dest

    if not output_dir:
        # if not output, then all content will be print to screen
        logger.info('event="No output_dir", will print content to screen"')
        return output_dir

    if os.path.exists(output_dir):
        logger.info('event="output_dir=%s already exist"',
                    output_dir)
        output_dir = os.path.join(output_dir,
                                  envs["product_id"] + "_temp_output")
        logger.info('event="generate a new output_dir=%s"', output_dir)
        if os.path.exists(output_dir):
            shutil.rmtree(output_dir)

    try:
        # copy file
        logger.info('event="Copying directory tree: src=%s dest=%s"',
                    src, output_dir)
        shutil.copytree(src, output_dir)

        # process each file's name
        logger.info('event="Replace each file name\'s placeholder under dir=%s"',
                    output_dir)
        move_list = []
        for dirName, subdirList, fileList in os.walk(output_dir):
            move_list.extend(check_file_list(dirName, subdirList, envs))
            move_list.extend(check_file_list(dirName, fileList, envs))

        for x, y, z in move_list:
            shutil.move(os.path.sep.join([x, y]), os.path.sep.join([x, z]))

    except Exception as e:
        if output_dir != dest and os.path.exists(output_dir):
            logger.info('clean temp_output_dir="%s"', output_dir)
            shutil.rmtree(output_dir)
        raise e
    return output_dir


def generate_alerts(src, dest, logger, envs, process_list=None,
                          skip_list=None):
    process_list = process_list or []
    skip_list = skip_list or []
    output_dir = dest
    package_dir = None
    output_content = {}
    conf_return = None
    html_return = None
    py_return = None
    global_settings = envs["global_settings"]

    try:
        if dest:
            output_dir = prepare_ta_directory_tree(src, dest, logger, envs)
            package_dir = os.path.join(dest, envs["product_id"])

        build_components = envs["build_components"]
        if build_components["conf"]:
            conf_return = generate_alert_actions_conf(
                input_setting=envs["schema.content"],
                package_path=package_dir,
                logger=logger,
                global_settings=global_settings)

        if build_components["html"]:
            html_return = generate_alert_actions_html_files(
                input_setting=envs["schema.content"],
                package_path=package_dir,
                logger=logger,
                html_setting=envs["html_setting"])

        if build_components["py"]:
            py_return = generate_alert_actions_py_files(
                input_setting=envs["schema.content"],
                package_path=package_dir,
                logger=logger,
                global_settings=global_settings
            )

        if conf_return:
            output_content["conf"] = conf_return
        if html_return:
            output_content["html"] = conf_return
        if py_return:
            output_content["py"] = py_return

        if output_dir != dest:
            """
            Which means the previous output_dir already there
            """
            from . import alert_actions_merge
            alert_actions_merge.merge(
                os.path.join(output_dir, envs["product_id"]),
                os.path.join(dest, envs["product_id"]))
            logger.info('event="merged %s to %s', output_dir, dest)
    finally:
        if output_dir != dest and os.path.exists(output_dir):
            logger.info('clean temp_output_dir="%s"', output_dir)
            shutil.rmtree(output_dir)

    return output_content


# SPDX-FileCopyrightText: 2020 2020
#
# SPDX-License-Identifier: Apache-2.0

"""
REST Builder.
"""

from __future__ import absolute_import

import collections
from splunktaucclib.rest_handler.schema import RestSchema

from .builder import (
    RestBuilder,
    RestBuilderError,
)

__all__ = [
    "RestBuilder",
    "RestBuilderError",
    "RestHandlerClass",
    "build",
]

__version__ = "0.0.0"


RestHandlerClass = collections.namedtuple("RestHandlerClass", ("module", "name"),)


def build(schema, handler, output_path, j2_env, post_process=None, *args, **kwargs):
    """
    Build REST for Add-on.

    :param schema: REST schema.
    :type schema: RestSchema
    :param handler: REST handler class import path:
            ``module.sub_module.RestHandlerClass``.
        The HandlerClass must be subclass of
        splunktaucclib.rest_handler.admin_external.AdminExternalHandler.
    :type handler: str
    :param output_path: path for output.
    :param post_process:
    :param args: args for post_process.
    :param kwargs: kwargs for post_process.
    :return:
    """

    def _parse_handler(handler_path):
        parts = handler_path.split(".")
        if len(parts) <= 1:
            raise RestBuilderError(
                "Invalid handler specified. "
                'It should be in form "module.sub_module.RestHandlerClass".'
            )
        return RestHandlerClass(module=".".join(parts[:-1]), name=parts[-1],)

    builder_obj = RestBuilder(schema, _parse_handler(handler), output_path)
    builder_obj.build()
    if post_process is not None:
        post_process(builder_obj, schema, *args, **kwargs)
    return builder_obj



# SPDX-FileCopyrightText: 2020 2020
#
# SPDX-License-Identifier: Apache-2.0

"""
This class is used to override the Python built-in ConfigParser, because
TA builder needs to support:
1. Read/Write conf files with comments
2. Additional comment prefix such as *
3. Support multiline end with \

There are a lot of legacy codes here. If you want to see our changes,
please search: ######## tab_update ########
"""

from future import standard_library
standard_library.install_aliases()
from builtins import str
import configparser

try:
    from collections import OrderedDict as _default_dict
except ImportError:
    # fallback for setup.py which hasn't yet built _collections
    _default_dict = dict


COMMENT_PREFIX = ";#*"
COMMENT_KEY = "__COMMENTS__"

class TABConfigParser(configparser.RawConfigParser):

    def _read(self, fp, fpname):
        """
        Override the built-in _read() method to read comments
        """
        from configparser import DEFAULTSECT, MissingSectionHeaderError, ParsingError

        cursect = None                        # None, or a dictionary
        optname = None
        lineno = 0
        e = None                              # None, or an exception

        ######## tab_update ########
        comment_index = 0
        self.top_comments = []
        self.fields_outside_stanza = []
        add_space_to_next_line = False
        ############################

        while True:
            line = fp.readline()
            if not line:
                break
            lineno = lineno + 1
            line = line.strip(" ")
            # comment or blank line?
            if line.strip() == '' or line[0] in COMMENT_PREFIX:

            ######## tab_update ########
            # save the lineno & comments
                if cursect:
                    name = "{}{}".format(COMMENT_KEY, comment_index)
                    comment_index += 1
                    cursect[name] = line
                else:
                    self.top_comments.append(line)
                continue
            ############################

            if line.split(None, 1)[0].lower() == 'rem' and line[0] in "rR":
                # no leading whitespace
                continue
            # continuation line?

            ######## tab_update ########
            # support multiline with \
            if add_space_to_next_line:
                line = " " + line

            if line.strip().endswith("\\"):
                line = line.rstrip("\\ ")
                add_space_to_next_line = True
            else:
                add_space_to_next_line = False
            ############################

            if line[0].isspace() and cursect is not None and optname:
                value = line.strip()
                if value:
                    cursect[optname].append(value)
            # a section header or option header?
            else:
                # is it a section header?
                mo = self.SECTCRE.match(line)
                if mo:
                    sectname = mo.group('header')
                    if sectname in self._sections:
                        cursect = self._sections[sectname]
                    elif sectname == DEFAULTSECT:
                        cursect = self._defaults
                    else:
                        cursect = self._dict()
                        cursect['__name__'] = sectname
                        self._sections[sectname] = cursect
                    # So sections can't start with a continuation line
                    optname = None
                # no section header in the file?

                elif cursect is None:
                    ######## tab_update ########
                    # disable the exception since splunk allows the field outside stanzas
#                     raise MissingSectionHeaderError(fpname, lineno, line)
                    self.fields_outside_stanza.append(line)
                ############################
                # an option line?
                else:
                    mo = self._optcre.match(line)
                    if mo:
                        optname, vi, optval = mo.group('option', 'vi', 'value')
                        optname = self.optionxform(optname.rstrip())
                        # This check is fine because the OPTCRE cannot
                        # match if it would set optval to None
                        if optval is not None:
                            if vi in ('=', ':') and ';' in optval:
                                # ';' is a comment delimiter only if it follows
                                # a spacing character
                                pos = optval.find(';')
                                if pos != -1 and optval[pos-1].isspace():
                                    optval = optval[:pos]
                            optval = optval.strip()
                            # allow empty values
                            if optval == '""':
                                optval = ''
                            cursect[optname] = [optval]
                        else:
                            # valueless option handling
                            cursect[optname] = optval
                    else:
                        # a non-fatal parsing error occurred.  set up the
                        # exception but keep going. the exception will be
                        # raised at the end of the file and will contain a
                        # list of all bogus lines
                        if not e:
                            e = ParsingError(fpname)
                        e.append(lineno, repr(line))
        # if any parsing errors occurred, raise an exception
        if e:
            raise e

        # join the multi-line values collected while reading
        all_sections = [self._defaults]
        all_sections.extend(list(self._sections.values()))
        for options in all_sections:
            for name, val in list(options.items()):
                if isinstance(val, list):
                    options[name] = '\n'.join(val)


    def write(self, fp):
        """
        Override the write() method to write comments
        """
        DEFAULTSECT = "DEFAULT"

        ######## tab_update ########
        if hasattr(self, "top_comments"):
            for comment in self.top_comments:
                fp.write(comment)

        if hasattr(self, "fields_outside_stanza"):
            for field in self.fields_outside_stanza:
                fp.write(field)
        ############################

        if self._defaults:
            fp.write("[%s]\n" % DEFAULTSECT)
            for (key, value) in list(self._defaults.items()):
                fp.write("%s = %s\n" % (key, str(value).replace('\n', '\n\t')))
            fp.write("\n")
        for section in self._sections:
            fp.write("[%s]\n" % section)
            for (key, value) in list(self._sections[section].items()):
                if key == "__name__":
                    continue

                ######## tab_update ########
                if key.startswith(COMMENT_KEY):
                    # only write the non empty line
                    if len(value.strip()) > 0:
                        fp.write(value)
                    # should continue as long as it is a comment line
                    continue
                ############################

                if (value is not None) or (self._optcre == self.OPTCRE):
                    key = " = ".join((key, str(value).replace('\n', '\n\t')))
                fp.write("%s\n" % (key))
            ######## tab_update ########
            # write the seperator line for stanza
            fp.write("\n")
            ############################

    def optionxform(self, optionstr):
        return optionstr

    def items(self, section):
        """
        Override the items() method to filter out the comments
        """
        items = configparser.RawConfigParser.items(self, section)

        res = []
        for k,v in items:
            if k.startswith(COMMENT_KEY):
                continue
            res.append((k,v))
        return res

    def options(self, section):
        options = configparser.RawConfigParser.options(self, section)

        res = []
        for opt in options:
            if opt.startswith(COMMENT_KEY):
                continue
            res.append(opt)

        return res

    def item_dict(self):
        res = {}
        sections = dict(self._sections)
        for section, key_values in list(sections.items()):
            kv = {}
            for k,v in list(key_values.items()):
                if not isinstance(k, str) or k.startswith(COMMENT_KEY) or k == "__name__":
                    continue
                kv[k] = v
            if kv:
                res[section] = kv
        return res



# SPDX-FileCopyrightText: 2020 2020
#
# SPDX-License-Identifier: Apache-2.0

from builtins import object
import json
import logging
import logging.handlers
import re
import splunk.rest as rest
from splunk.util import normalizeBoolean

class InvalidResultID(Exception):
    pass

class ModularAction(object):
    DEFAULT_MESSAGE = 'sendmodaction - signature="%s" action_name="%s" search_name="%s" sid="%s" orig_sid="%s" rid="%s" orig_rid="%s" app="%s" user="%s" action_mode="%s" action_status="%s"'

    ## we require a logging instance
    '''
2015-03-07T01:41:42.430696 got arguments ['/opt/splunk/etc/apps/logger_app/bin/logger.py', '--execute']
2015-03-07T01:41:42.430718 got payload: <?xml version="1.0" encoding="UTF-8"?>
<alert>
  <app> logger_app </app>
  <owner>admin</owner>
  <results_file>/opt/splunk/var/run/splunk/dispatch/rt_scheduler__admin__ logger_app__RMD5910195c23186c103_at_1425692383_0.0/results.csv.gz</results_file>
  <results_link>http://myserver:8000/app/logger_app/@go?sid=rt_scheduler__admin__ logger_app__RMD5910195c23186c103_at_1425692383_0.0</results_link>
  <server_host>myserver</server_host>
  <server_uri>https://127.0.0.1:8089</server_uri>
  <session_key>OCmOZHf37O^9fDktTrvNc6Kidz^68zs0Y7scufwRo6Lpdi5ZGmtxsPbIUlUKtjt9ZPG7gKz4Dq8_eVntQ5EGR^N9rqkmg1dREAp8FFCduDwwvl6pEXEB^4w3MS6suwp9acw7JOlb</session_key>
  <sid>rt_scheduler__admin__ logger_app__RMD5910195c23186c103_at_1425692383_0.0</sid>
  <search_name>my_saved_search</search_name>
  <configuration>
     <stanza name= my_saved_search"/>
  </configuration>
</alert>

json format
{
u'results_file': u'/Users/zhanghong/splunk_env/splunk_env_6.3.3/splunk/var/run/splunk/dispatch/scheduler__admin__search__RMD5f2f82c6b0b712cbb_at_1469423700_4185/per_result_alert/tmp_1.csv.gz',
u'server_host': u'hozhang-mbpo.sv.splunk.com',
u'sid': u'scheduler__admin__search__RMD5f2f82c6b0b712cbb_at_1469423700_4185',
u'result': {u'_bkt': u'main~899~DC01C3FC-2111-4BA8-BA2E-B6F07C5C9C30', u'gid': u'00g17z5bhl0Dpv2FF1d8', u'splunk_server': u'hozhang-mbpo.sv.splunk.com', u'_time': u'1467171788', u'_raw': u'demo_case=101 uid=00u17z584iaBW2Biz1d8 gid=00g17z5bhl0Dpv2FF1d8', u'demo_case': u'101', u'uid': u'00u17z584iaBW2Biz1d8', u'_serial': u'1', u'tag::eventtype': u'', u'_si': [u'hozhang-mbpo.sv.splunk.com', u'main'], u'index': u'main', u'sourcetype': u'okta_demo', u'uname': u'', u'eventtype': u'', u'_kv': u'1', u'host': u'hozhang-mbpo.sv.splunk.com', u'_sourcetype': u'okta_demo', u'punct': u'=_=_=', u'gname': u'', u'_indextime': u'1467337748', u'splunk_server_group': u'', u'tag': u'', u'linecount': u'1', u'_cd': u'899:181', u'product': u'', u'vendor': u'', u'ids_type': u'', u'source': u'/Users/zhanghong/Downloads/okta_demo_cases', u'timestamp': u'none'}, u'configuration': {u'group_id': u'00g17z5bhl0Dpv2FF1d8', u'action': u'add', u'user_id': u'00u17z584iaBW2Biz1d8'},
u'owner': u'admin',
u'results_link': u'http://hozhang-mbpo.sv.splunk.com:8000/app/search/search?q=%7Cloadjob%20scheduler__admin__search__RM...',
u'session_key': u'e6w4SwfYg5BLuXZ6u78^TMWz1rdMGzSbpIWD8kv0SzFIyuuCt_qk2Lxv1zOzC4vue7OaWGVggaLJN5tyn_uQSE_rLm^l3Z0TA8UNmX1TamWTDsAR9kdtCL67Vz0trb3wfMrWFkYOR3mUur0',
u'app': u'search',
u'server_uri': u'https://127.0.0.1:8089',
u'search_name': u'demo_okta_alert_002'
}
    '''
    def __init__(self, settings, logger, action_name='unknown'):
        self.settings      = json.loads(settings)
        self.logger        = logger
        self.session_key   = self.settings.get('session_key')
        self.sid           = self.settings.get('sid')
        ## if sid contains rt_scheduler with snapshot-sid; drop snapshot-sid
        ## sometimes self.sid may be an integer (1465593470.1228)
        try:
            rtsid    = re.match('^(rt_scheduler.*)\.\d+$', self.sid)
            if rtsid:
                self.sid = rtsid.group(1)
        except:
            pass
        self.orig_sid      = ''
        self.rid           = ''
        self.orig_rid      = ''
        self.results_file  = self.settings.get('results_file')
        self.search_name   = self.settings.get('search_name')
        self.app           = self.settings.get('app')
        self.user          = self.settings.get('user') or self.settings.get('owner')
        self.configuration = self.settings.get('configuration', {})
        ## enforce configuration is a 'dict'
        if not isinstance(self.configuration, dict):
            self.configuration = {}
        ## set loglevel to DEBUG if verbose
        if normalizeBoolean(self.configuration.get('verbose', 'false')):
            self.logger.setLevel(logging.DEBUG)
            self.logger.debug("loglevel set to DEBUG")
        ## use | sendalert param.action_name=$action_name$
        self.action_name = self.configuration.get('action_name') or action_name
        ## use search_name to determine action_mode
        if self.search_name:
            self.action_mode = 'saved'
        else:
            self.action_mode = 'adhoc'

        self.action_status = ''
        ## Since we don't use the result object we get from settings it will be purged
        try:
            del self.settings['result']
        except Exception:
            pass

    ## The purpose of this method is to populate the job variable with the contents from REST (/services/search/jobs/<sid>)
    ## SPL-112815 - sendalert - not all $job.<param>$ parameters come through
    def addjobinfo(self):
        self.job = {}
        if self.sid:
            try:
                response, content = rest.simpleRequest('search/jobs/%s' % self.sid, sessionKey=self.session_key, getargs={'output_mode': 'json'})
                if response.status == 200:
                    self.job = json.loads(content)['entry'][0]['content']
                    self.logger.info(self.message('Successfully retrieved search job info'))
                    self.logger.debug(self.job)
                else:
                    self.logger.warn(self.message('Could not retrieve search job info'))
            except Exception as e:
                self.logger.warn(self.message('Could not retrieve search job info'))

    ## The purpose of this method is to provide a common messaging interface
    def message(self, signature, status=None):
        status  = status or self.action_status or ''
        message = ModularAction.DEFAULT_MESSAGE % (signature or '', self.action_name or '', self.search_name or '', self.sid or '', self.orig_sid or '', self.rid or '', self.orig_rid or '', self.app or '', self.user or '', self.action_mode or '', status)
        ## prune empty string key-value pairs
        for match in re.finditer('[A-Za-z_]+=\"\"(\s|$)', message):
            message = message.replace(match.group(0),'',1)
        return message.rstrip()

    ## The purpose of this method is to update per-result ModAction attributes
    def update(self, result):
        ## This is for events/results that were created as the result of a previous action
        self.orig_sid = result.get('orig_sid', '')
        ## This is for events/results that were created as the result of a previous action
        self.orig_rid = result.get('orig_rid', '')
        if 'rid' in result:
            self.rid = result['rid']
        else:
            raise InvalidResultID('Result must have an ID')

    ## The purpose of this method is to generate per-result invocation messages
    def invoke(self):
        self.logger.info(self.message('Invoking modular action'))

    def dowork(self):
        return
